/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009, 2011
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009, 2011
+	Copyright Daan Vreeken <Daan@Vitsch.nl> - Vitsch Electronics, 2010

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdlib.h>
#include <string.h> 

#include <controller/controller_block.h>

#include "block_beckhoff_el2xxx.h"
#include "ec.h"
#include "esc.h"
#include "log.h"

struct controller_block_private {
	bool	*output1;
	bool	*output2;
	bool	*output3;
	bool	*output4;
	bool	*output5;
	bool	*output6;
	bool	*output7;
	bool	*output8;
	
	unsigned char *tx_buffer;
	int 	outputs;
};

static void calculate_8(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	
	priv->tx_buffer[0] =
		*priv->output1 |
		(*priv->output2 << 1) |
		(*priv->output3 << 2) |
		(*priv->output4 << 3) |
		(*priv->output5 << 4) |
		(*priv->output6 << 5) |
		(*priv->output7 << 6) |
		(*priv->output8 << 7);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_4(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	
	priv->tx_buffer[0] =
		*priv->output1 |
		(*priv->output2 << 1) |
		(*priv->output3 << 2) |
		(*priv->output4 << 3);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_2(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	
	priv->tx_buffer[0] =
		*priv->output1 |
		(*priv->output2 << 1);

	ec_tx_pdo(controller_samplenr);
}

struct beckhoff_el2xxx_type {
	char *name;
	uint32_t productcode;
	int outputs;
	struct controller_block_interm_list *interms;
};

static struct controller_block_interm_list interms_2[] = {
	{ "output1",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output1)  },
	{ "output2",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output2)  },
	{ NULL },
};

static struct controller_block_interm_list interms_4[] = {
	{ "output1",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output1)  },
	{ "output2",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output2)  },
	{ "output3",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output3)  },
	{ "output4",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output4)  },
	{ NULL },
};

static struct controller_block_interm_list interms_8[] = {
	{ "output1",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output1)  },
	{ "output2",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output2)  },
	{ "output3",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output3)  },
	{ "output4",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output4)  },
	{ "output5",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output5)  },
	{ "output6",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output6)  },
	{ "output7",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output7)  },
	{ "output8",  CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output8)  },
	{ NULL },
};

static struct beckhoff_el2xxx_type devlist[] = {
	{ "EL2002", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2002, 2, interms_2 },
	{ "EL2004", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2004, 4, interms_4 },
	{ "EL2008", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2008, 8, interms_8 },
	{ "EL2022", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2022, 2, interms_2 },
	{ "EL2024", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2024, 4, interms_4 },
	{ "EL2032", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2032, 2, interms_2 },
	{ "EL2034", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2034, 4, interms_4 },
	{ "EL2042", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2042, 2, interms_2 },
	{ "EL2084", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2084, 4, interms_4 },
	{ "EL2088", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2088, 8, interms_8 },
	{ "EL2124", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2124, 4, interms_4 },
	{ "EL2202", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2202, 2, interms_2 },
	{ "EL2252", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2252, 2, interms_2 },
	{ "EL2262", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2262, 2, interms_2 },
	{ "EL2602", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2602, 2, interms_2 },
	{ "EL2612", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2612, 2, interms_2 },
	{ "EL2622", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2622, 2, interms_2 },
	{ "EL2624", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2624, 4, interms_4 },
	{ "EL2712", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2712, 2, interms_2 },
	{ "EL2722", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2722, 2, interms_2 },
	{ "EL2732", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2732, 2, interms_2 },
};

struct controller_block * block_beckhoff_el2xxx_create(char *name, va_list ap)
{
	struct controller_block *block;
	struct ec_dgram_addr addr;
	struct esc_syncmanager_info pdo_tx;
	int slaves;
	int i;
	bool found;
	int dev = -1;
	char *eltype;
	int devno;
	bool sim = !ec_initialized();
	struct timespec timeout = { 1, 0 };
		
	eltype = va_arg(ap, char *);
	devno = va_arg(ap, int);
	
	for (i = 0; i < sizeof(devlist)/sizeof(struct beckhoff_el2xxx_type); i++) {
		if (!strcmp(eltype, devlist[i].name)) {
			dev = i;
			break;
		}
	}
	if (dev < 0)
		goto err_ethercat;
	
	if (!sim) {
		slaves = ec_slave_count();
		found = false;
		for (i = 0; i < slaves; i++) {
			struct ec_dgram_addr devaddr;
			uint32_t vendor, product;
			
			ec_addr_set_auto_inc_nr(&devaddr, i);
			vendor = esc_esi_vendorid_get(&devaddr);
			product = esc_esi_productcode_get(&devaddr);
			if (vendor == ESC_ESI_VENDORID_BECKHOFF &&
			    product == devlist[dev].productcode) {
				if (!devno || devno == i) {
					found = true;
					addr = devaddr;
				}
			}
		}
		if (!found) {
			log_send(LOG_T_ERROR, "%s: Device not found\n", name);
			goto err_ethercat;
		}
	}
	
	block = controller_block_alloc(
	    sim ? "beckhoff_el2xxx_sim" : "beckhoff_el2xxx",
	    name, sizeof(struct controller_block_private));

	if (!sim) {
		esc_init(&addr);

		pdo_tx.start = 0x0f00;
		pdo_tx.len = 0x1;
		pdo_tx.sm = 0;

		esc_pdo_tx_set(&addr, &pdo_tx, &block->private->tx_buffer, true);

		if (esc_al_state_set(&addr, ESC_AL_STATE_INIT, &timeout) < 0)
			log_send(LOG_T_ERROR, "Could not go to state init\n");
		if (esc_al_state_set(&addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout) < 0)
			log_send(LOG_T_ERROR, "Could not go to state pre operational\n");
		if (esc_al_state_set(&addr, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout) < 0)
			log_send(LOG_T_ERROR, "Could not go to state safe operational\n");
		if (esc_al_state_set(&addr, ESC_AL_STATE_OPERATIONAL, &timeout) < 0)
			log_send(LOG_T_ERROR, "Could not go to state safe operational\n");
		log_send(LOG_T_DEBUG, "state: %d\n", esc_al_state_get(&addr));
		log_send(LOG_T_DEBUG, "statuscode: %d\n", esc_al_status_code_get(&addr));
	}
	
	/* Outputs on the hardware are inputs on this software block */
	if (controller_block_interm_list_init(block, devlist[dev].interms))
		goto err_inputs;	
	
	switch (devlist[dev].outputs) {
		case 2:
			block->calculate = calculate_2;
			break;
		case 4:
			block->calculate = calculate_4;
			break;
		case 8:
			block->calculate = calculate_8;
			break;
	}
	if (sim)
		block->calculate = NULL;

	controller_block_add(block);
	return block;

err_inputs:
	controller_block_free(block);
err_ethercat:
	return NULL;
}
