/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>

#include "block_beckhoff_el4xxx.h"
#include "ec.h"
#include "esc.h"
#include "log.h"

struct controller_block_private {
	float	*output1;
	float	*output2;
	float	*output3;
	float	*output4;
	float	*output5;
	float	*output6;
	float	*output7;
	float	*output8;

	float gain;
	float offset;
	
	unsigned char *tx_buffer;
};

static void calculate_1u(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1;
	struct {
		uint16_t value1;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	if (value1 > 65535)
		value1 = 65535;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16(value1);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_1s(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1;
	struct {
		int16_t value1;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	if (value1 > 32767)
		value1 = 32767;
	if (value1 < -32768)
		value1 = -32768;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16((int16_t)value1);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_2u(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1, value2;
	struct {
		uint16_t value1;
		uint16_t value2;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	if (value1 > 65535)
		value1 = 65535;
	value2 = (priv->offset + *priv->output2) * priv->gain;
	if (value2 > 65535)
		value2 = 65535;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16(value1);
	out->value2 = htole16(value2);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_2s(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1, value2;
	struct {
		int16_t value1;
		int16_t value2;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	value2 = (priv->offset + *priv->output2) * priv->gain;
	if (value1 > 32767)
		value1 = 32767;
	if (value1 < -32768)
		value1 = -32768;
	if (value2 > 32767)
		value2 = 32767;
	if (value2 < -32768)
		value2 = -32768;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16((int16_t)value1);
	out->value2 = htole16((int16_t)value2);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_4u(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1, value2, value3, value4;
	struct {
		uint16_t value1;
		uint16_t value2;
		uint16_t value3;
		uint16_t value4;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	if (value1 > 65535)
		value1 = 65535;
	value2 = (priv->offset + *priv->output2) * priv->gain;
	if (value2 > 65535)
		value2 = 65535;
	value3 = (priv->offset + *priv->output3) * priv->gain;
	if (value3 > 65535)
		value3 = 65535;
	value4 = (priv->offset + *priv->output4) * priv->gain;
	if (value4 > 65535)
		value4 = 65535;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16(value1);
	out->value2 = htole16(value2);
	out->value3 = htole16(value3);
	out->value4 = htole16(value4);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_4s(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1, value2, value3, value4;
	struct {
		int16_t value1;
		int16_t value2;
		int16_t value3;
		int16_t value4;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	value2 = (priv->offset + *priv->output2) * priv->gain;
	value3 = (priv->offset + *priv->output3) * priv->gain;
	value4 = (priv->offset + *priv->output4) * priv->gain;
	if (value1 > 32767)
		value1 = 32767;
	if (value1 < -32768)
		value1 = -32768;
	if (value2 > 32767)
		value2 = 32767;
	if (value2 < -32768)
		value2 = -32768;
	if (value3 > 32767)
		value3 = 32767;
	if (value3 < -32768)
		value3 = -32768;
	if (value4 > 32767)
		value4 = 32767;
	if (value4 < -32768)
		value4 = -32768;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16((int16_t)value1);
	out->value2 = htole16((int16_t)value2);
	out->value3 = htole16((int16_t)value3);
	out->value4 = htole16((int16_t)value4);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_8u(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1, value2, value3, value4;
	float value5, value6, value7, value8;
	struct {
		uint16_t value1;
		uint16_t value2;
		uint16_t value3;
		uint16_t value4;
		uint16_t value5;
		uint16_t value6;
		uint16_t value7;
		uint16_t value8;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	if (value1 > 65535)
		value1 = 65535;
	value2 = (priv->offset + *priv->output2) * priv->gain;
	if (value2 > 65535)
		value2 = 65535;
	value3 = (priv->offset + *priv->output3) * priv->gain;
	if (value3 > 65535)
		value3 = 65535;
	value4 = (priv->offset + *priv->output4) * priv->gain;
	if (value4 > 65535)
		value4 = 65535;
	value5 = (priv->offset + *priv->output5) * priv->gain;
	if (value5 > 65535)
		value5 = 65535;
	value6 = (priv->offset + *priv->output6) * priv->gain;
	if (value6 > 65535)
		value6 = 65535;
	value7 = (priv->offset + *priv->output7) * priv->gain;
	if (value7 > 65535)
		value7 = 65535;
	value8 = (priv->offset + *priv->output8) * priv->gain;
	if (value8 > 65535)
		value8 = 65535;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16(value1);
	out->value2 = htole16(value2);
	out->value3 = htole16(value3);
	out->value4 = htole16(value4);
	out->value5 = htole16(value5);
	out->value6 = htole16(value6);
	out->value7 = htole16(value7);
	out->value8 = htole16(value8);

	ec_tx_pdo(controller_samplenr);
}

static void calculate_8s(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float value1, value2, value3, value4;
	float value5, value6, value7, value8;
	struct {
		int16_t value1;
		int16_t value2;
		int16_t value3;
		int16_t value4;
		int16_t value5;
		int16_t value6;
		int16_t value7;
		int16_t value8;
	} __attribute__((packed)) *out;
	
	value1 = (priv->offset + *priv->output1) * priv->gain;
	value2 = (priv->offset + *priv->output2) * priv->gain;
	value3 = (priv->offset + *priv->output3) * priv->gain;
	value4 = (priv->offset + *priv->output4) * priv->gain;
	value5 = (priv->offset + *priv->output5) * priv->gain;
	value6 = (priv->offset + *priv->output6) * priv->gain;
	value7 = (priv->offset + *priv->output7) * priv->gain;
	value8 = (priv->offset + *priv->output8) * priv->gain;
	if (value1 > 32767)
		value1 = 32767;
	if (value1 < -32768)
		value1 = -32768;
	if (value2 > 32767)
		value2 = 32767;
	if (value2 < -32768)
		value2 = -32768;
	if (value3 > 32767)
		value3 = 32767;
	if (value3 < -32768)
		value3 = -32768;
	if (value4 > 32767)
		value4 = 32767;
	if (value4 < -32768)
		value4 = -32768;
	if (value5 > 32767)
		value5 = 32767;
	if (value5 < -32768)
		value5 = -32768;
	if (value6 > 32767)
		value6 = 32767;
	if (value6 < -32768)
		value6 = -32768;
	if (value7 > 32767)
		value7 = 32767;
	if (value7 < -32768)
		value7 = -32768;
	if (value8 > 32767)
		value8 = 32767;
	if (value8 < -32768)
		value8 = -32768;
	
	out = (void*)priv->tx_buffer;
	
	out->value1 = htole16((int16_t)value1);
	out->value2 = htole16((int16_t)value2);
	out->value3 = htole16((int16_t)value3);
	out->value4 = htole16((int16_t)value4);
	out->value5 = htole16((int16_t)value5);
	out->value6 = htole16((int16_t)value6);
	out->value7 = htole16((int16_t)value7);
	out->value8 = htole16((int16_t)value8);

	ec_tx_pdo(controller_samplenr);
}

struct beckhoff_el4xxx_type {
	char *name;
	uint32_t productcode;
	int outputs;
	float gain;
	float offset;
	bool sign;
};

static struct beckhoff_el4xxx_type devlist[] = {
	{ "EL4001", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4001, 1, 4096/10.0,  0.0,   false },
	{ "EL4002", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4002, 2, 4096/10.0,  0.0,   false },
	{ "EL4004", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4004, 4, 4096/10.0,  0.0,   false },
	{ "EL4008", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4008, 8, 4096/10.0,  0.0,   false },
	{ "EL4011", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4001, 1, 4096/0.02,  0.0,   false },
	{ "EL4012", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4002, 2, 4096/0.02,  0.0,   false },
	{ "EL4014", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4004, 4, 4096/0.02,  0.0,   false },
	{ "EL4018", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4008, 8, 4096/0.02,  0.0,   false },
	{ "EL4021", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4001, 1, 4096/0.016, -0.004, false },
	{ "EL4022", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4002, 2, 4096/0.016, -0.004, false },
	{ "EL4024", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4004, 4, 4096/0.016, -0.004, false },
	{ "EL4028", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4008, 8, 4096/0.016, -0.004, false },
	{ "EL4001", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4001, 1, 2048/10.0,  0.0,   true  },
	{ "EL4002", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4002, 2, 2048/10.0,  0.0,   true  },
	{ "EL4004", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4004, 4, 2048/10.0,  0.0,   true  },
	{ "EL4008", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4008, 8, 2048/10.0,  0.0,   true  },
	{ "EL4102", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4102, 2, 65536/10.0, 0.0,   false },
	{ "EL4112", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4112, 2, 65536/0.02, 0.0,   false },
	{ "EL4114", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4114, 4, 65536/0.02, 0.0,   false },
	{ "EL4122", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4122, 2, 65536/0.016, 0.0,  false },
	{ "EL4132", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4132, 2, 32768/10.0, 0.0,   true  },
	{ "EL4134", ESC_ESI_PRODUCTCODE_BECKHOFF_EL4134, 4, 32768/10.0, 0.0,   true  },
};

struct controller_block * block_beckhoff_el4xxx_create(char *name, va_list ap)
{
	struct controller_block *block;
	struct ec_dgram_addr addr;
	struct esc_syncmanager_info pdo_tx, pdo_rx;
	int slaves;
	int i;
	bool found;
	int dev = -1;
	char *eltype;
	int devno;
	uint8_t val8;
	uint16_t leval16;
	struct timespec timeout = { 1, 0 };
		
	struct esc_mailbox *mb;
	struct esc_syncmanager_info mb_rd, mb_wr;

	eltype = va_arg(ap, char *);
	devno = va_arg(ap, int);
	
	for (i = 0; i < sizeof(devlist)/sizeof(struct beckhoff_el4xxx_type); i++) {
		if (!strcmp(eltype, devlist[i].name)) {
			dev = i;
			break;
		}
	}
	if (dev < 0)
		return NULL;
	
	slaves = ec_slave_count();
	found = false;
	for (i = 0; i < slaves; i++) {
		struct ec_dgram_addr devaddr;
		uint32_t vendor, product;
		
		ec_addr_set_auto_inc_nr(&devaddr, i);
		vendor = esc_esi_vendorid_get(&devaddr);
		product = esc_esi_productcode_get(&devaddr);
		if (vendor == ESC_ESI_VENDORID_BECKHOFF &&
		    product == devlist[dev].productcode) {
			if (!devno || devno == i) {
				found = true;
				addr = devaddr;
			}
		}
	}
	if (!found) {
		log_send(LOG_T_ERROR, "%s: Device not found\n", name);
		return NULL;
	}
	
	block = malloc(sizeof(struct controller_block));
	if (!block)
		return NULL;

	block->private = malloc(sizeof(struct controller_block_private));
	if (!block->private)
		return NULL;

	block->type = "beckhoff_el4xxx";
	block->name = strdup(name);

	esc_init(&addr);

	pdo_tx.start = 0x1000;
	pdo_tx.len = 0x2 * devlist[dev].outputs;
	pdo_tx.sm = 2;
	

	pdo_rx.start = 0x1100;
	pdo_rx.len = 0;
	pdo_rx.sm = 3;

	esc_pdo_tx_set(&addr, &pdo_tx, &block->private->tx_buffer, false);
	esc_pdo_rx_set(&addr, &pdo_rx, NULL);

	if (esc_al_state_set(&addr, ESC_AL_STATE_INIT, &timeout) < 0)
		printf("Could not go to state init\n");

	/* mailbox read (slave->master) @0x18f6 */
	mb_rd.start = 0x18f6;
	mb_rd.len = 246;
	mb_rd.sm = 1;
	
	/* mailbox write (master->slave) @0x1800 */
	mb_wr.start = 0x1800;
	mb_wr.len = 246;
	mb_wr.sm = 0;
	
	mb = esc_mailbox_create(&addr, &mb_rd, &mb_wr);
	esc_coe_create(mb);

	if (esc_al_state_set(&addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state pre operational\n");

	esc_dc_init(&addr);
	esc_dc_sync0_cycle_time(&addr, 1000*1000);

leval16 = 0x4;
addr.addr.position.off = 0x604 + 16 * 0;
ec_datagram_write(&addr, &leval16, 2);
leval16 = 0x1000;
addr.addr.position.off = 0x608 + 16 * 0;
ec_datagram_write(&addr, &leval16, 2);
val8 = 3;
addr.addr.position.off = 0x60b + 16 * 0;
ec_datagram_write(&addr, &val8, 1);
val8 = 1;
addr.addr.position.off = 0x60c + 16 * 0;
ec_datagram_write(&addr, &val8, 1);

leval16 = 0x4;
addr.addr.position.off = 0x604 + 16 * 1;
ec_datagram_write(&addr, &leval16, 2);
leval16 = 0x1100;
addr.addr.position.off = 0x608 + 16 * 1;
ec_datagram_write(&addr, &leval16, 2);
val8 = 3;
addr.addr.position.off = 0x60b + 16 * 1;
ec_datagram_write(&addr, &val8, 1);
val8 = 1;
addr.addr.position.off = 0x60c + 16 * 1;
ec_datagram_write(&addr, &val8, 1);


	addr.addr.position.off = 0x981;
	val8 = 0x00;
	ec_datagram_write(&addr, &val8, 1);


addr.addr.position.off = 0x420;
leval16 = 0x0000;
ec_datagram_write(&addr, &leval16, 2);


addr.addr.position.off = 0x442;
val8 = 0;
ec_datagram_write(&addr, &val8, 1);


	if (esc_al_state_set(&addr, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state safe operational\n");

	if (esc_al_state_set(&addr, ESC_AL_STATE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state safe operational\n");
	printf("state: %d\n", esc_al_state_get(&addr));
	printf("statuscode: %d\n", esc_al_status_code_get(&addr));

	block->outputs = 0;
	block->output = NULL;
	
	/* Outputs on the hardware are inputs on this software block */
	block->inputs = devlist[dev].outputs;
	block->input = malloc(sizeof(struct controller_block_interm) * 8);
	if (!block->input)
		return NULL;
	block->input[0].name = "output1";
	block->input[0].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[0].value.f = &block->private->output1;
	block->input[0].ghostof = NULL;
	block->input[1].name = "output2";
	block->input[1].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[1].value.f = &block->private->output2;
	block->input[1].ghostof = NULL;
	block->input[2].name = "output3";
	block->input[2].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[2].value.f = &block->private->output3;
	block->input[2].ghostof = NULL;
	block->input[3].name = "output4";
	block->input[3].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[3].value.f = &block->private->output4;
	block->input[3].ghostof = NULL;
	block->input[4].name = "output5";
	block->input[4].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[4].value.f = &block->private->output5;
	block->input[4].ghostof = NULL;
	block->input[5].name = "output6";
	block->input[5].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[5].value.f = &block->private->output6;
	block->input[5].ghostof = NULL;
	block->input[6].name = "output7";
	block->input[6].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[6].value.f = &block->private->output7;
	block->input[6].ghostof = NULL;
	block->input[7].name = "output8";
	block->input[7].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[7].value.f = &block->private->output8;
	block->input[7].ghostof = NULL;
	
	switch (devlist[dev].outputs) {
		case 1:
			if (devlist[dev].sign)
				block->calculate = calculate_1s;
			else
				block->calculate = calculate_1u;
			break;
		case 2:
			if (devlist[dev].sign)
				block->calculate = calculate_2s;
			else
				block->calculate = calculate_2u;
			break;
		case 4:
			if (devlist[dev].sign)
				block->calculate = calculate_4s;
			else
				block->calculate = calculate_4u;
			break;
		case 8:
			if (devlist[dev].sign)
				block->calculate = calculate_8s;
			else
				block->calculate = calculate_8u;
			break;
	}
	block->private->gain = devlist[dev].gain;
	block->private->offset = devlist[dev].offset;
	
	block->params = 0;
	block->param = NULL;
	block->param_get = NULL;
	block->param_set = NULL;

	controller_block_add(block);
	return block;
}
