/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef _INCLUDE_CANOPEN_H_
#define _INCLUDE_CANOPEN_H_

#include <unistd.h>
#include <stdint.h>

struct canopen_dev {
	ssize_t (*send_request)(struct canopen_dev *dev, void *req, size_t len, 
	    struct timespec *timeout);
	ssize_t (*recv_response)(struct canopen_dev *dev, void *res, size_t len,
	    struct timespec *timeout);
	void (*emergency_handler)(struct canopen_dev *dev, uint16_t code, uint8_t reg, uint8_t vendor[5]);

	char *logname;

	pthread_mutex_t lock;

	struct canopen_dev_priv *priv;
};

ssize_t canopen_write_param(struct canopen_dev *dev,
    unsigned index, unsigned subindex,
    void *data, ssize_t bytes);
ssize_t canopen_read_param(struct canopen_dev *dev,
    unsigned index, unsigned subindex,
    void *data, ssize_t bytes);

void canopen_emergency_handler_attach(struct canopen_dev *dev,
    void (*func)(struct canopen_dev *dev, uint16_t code, uint8_t register, uint8_t vendor[5]));

int canopen_init(struct canopen_dev *dev);

void canopen_emergency_handler(struct canopen_dev *dev, void *msg, size_t len);

#endif /* _INCLUDE_CANOPEN_H_ */
