/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2008
	Copyright Stichting C.A. Muller Radioastronomiestation, 2008

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef _INCLUDE_EC_STOEBER_H_
#define _INCLUDE_EC_STOEBER_H_

#include "ec.h"
#include "canopen.h"

#define STOEBER_PARAM2INDEX(group, nr) (0x2000 + ((group) - 'A') * 0x200 + nr)

#define STOEBER_PDO_MAP(g,nr,s,l) \
    htole32((STOEBER_PARAM2INDEX(g,nr)<<16)|((s)<<8)|l)

struct ec_stoeber {
	char *dev;
	struct ec_dgram_addr addr;
	float max_speed;
	float max_accel;
	float max_torque;

	float standstill_torque;
	float brake_resistor;
	float brake_power;

	uint32_t serial_number;
	
	struct canopen_dev *canopen_dev;
	
	/* rx pdo */
	unsigned char *rx_buffer; /* 14 */
	struct {
		uint32_t I80;
		uint32_t E91;
		uint16_t E90;
		uint16_t E19;
		uint16_t E10;
	} rx;
	
	/* tx pdo */
	unsigned char *tx_buffer; /* 12 */;
	struct {
		uint8_t A180;
		uint8_t F210;
		uint16_t I210;
		uint32_t I213;
		uint16_t I215;
		uint16_t C230;
	} tx;
};

int ec_stoeber_init(struct ec_stoeber *stbr);

#endif /* _INCLUDE_EC_STOEBER_H_ */
