/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "ec.h"
#include "esc.h"
#include "canopen.h"

int main (int argc, char **argv)
{
	int ret;
	unsigned char buffer[4096];
	struct ec_dgram_addr addr;
	struct esc_mailbox *mb;
	struct canopen_dev *canopen_dev;
	time_t t;
	int i;
	int16_t speed = 0;
	struct timespec timeout = { 1, 0 };


	/*****************************************************************
		init
	 *****************************************************************/
	printf("\n----------\ninit\n----------\n");

	ret = ec_init("eth0", 0);
	printf("ec_init(): %d\n", ret);


	/*****************************************************************
		ESC read/write test
	 *****************************************************************/
	printf("\n----------\nESC read/write test\n----------\n");


	addr.type = EC_DGRAM_ADDR_AUTO_INC;
	addr.addr.position.pos = htole16(0x0000);
	addr.addr.position.off = htole16(0x0000);
	
	ret = ec_datagram_read(&addr, buffer, 20);
	printf("ec_datagram_read(): %d\n", ret);
	
	if (ret == 20) {
		int i;
		
		for (i = 0; i < 20; i++) {
			printf("%02x ", buffer[i]);
		}
		printf("\n");
	}
	
	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("esc_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}

	addr.type = EC_DGRAM_ADDR_AUTO_INC;
	addr.addr.position.pos = htole16(0xaaaa);
	addr.addr.position.off = htole16(0xbeef);

	ret = ec_datagram_write(&addr, buffer, 20);
	printf("ec_datagram_write(): %d\n", ret);

	/*****************************************************************
		init state test
	 *****************************************************************/
	printf("\n----------\ninit state test\n----------\n");

	addr.type = EC_DGRAM_ADDR_AUTO_INC;
	addr.addr.position.pos = htole16(0x0000);
	addr.addr.position.off = htole16(0x0000);
	
	printf("esc_al_state_get(): %d\n", esc_al_state_get(&addr));

	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("esc_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}
	
	printf("esc_al_state_set(1==ESC_AL_STATE_INIT): %d\n",
	    esc_al_state_set(&addr, ESC_AL_STATE_INIT, &timeout));

	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("esc_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}

	{
		/* sm0 mailbox write (master->slave) @0x1800*/
		struct ec_dgram_addr addrt;
		uint16_t val;

		addrt.type = EC_DGRAM_ADDR_AUTO_INC;
		addrt.addr.position.pos = 0x0000;
		
		addrt.addr.position.off = 0x0806;
		val = 0x00;
		ec_datagram_write(&addrt, &val, 1);

		addrt.addr.position.off = 0x0800;
		val = 0x1800;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0802;
//		val = 0x00f6;
		val = 0x0040;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0804;
		val = 0x26;
		ec_datagram_write(&addrt, &val, 1);
		addrt.addr.position.off = 0x0806;
		val = 0x01;
		ec_datagram_write(&addrt, &val, 1);

	}
	{
		/* sm1 mailbox read (slave->master) @0x18f6 */
		struct ec_dgram_addr addrt;
		uint16_t val;
		
		addrt.type = EC_DGRAM_ADDR_AUTO_INC;
		addrt.addr.position.pos = 0x0000;

		addrt.addr.position.off = 0x0806 + 8;
		val = 0x00;
		ec_datagram_write(&addrt, &val, 1);

		addrt.addr.position.off = 0x0800 + 8;
		val = 0x18f6;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0802 + 8;
//		val = 0x00f6;
		val = 0x0040;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0804 + 8;
		val = 0x22;
		ec_datagram_write(&addrt, &val, 1);
		addrt.addr.position.off = 0x0806 + 8;
		val = 0x01;
		ec_datagram_write(&addrt, &val, 1);

	}
	{
		/* sm2 tx pdo (master->slave) @0x1600 */
		struct ec_dgram_addr addrt;
		uint16_t val;
		
		addrt.type = EC_DGRAM_ADDR_AUTO_INC;
		addrt.addr.position.pos = 0x0000;

		addrt.addr.position.off = 0x0806 + 16;
		val = 0x00;
		ec_datagram_write(&addrt, &val, 1);

		addrt.addr.position.off = 0x0800 + 16;
		val = 0x1600;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0802 + 16;
		val = 0x000a;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0804 + 16;
		val = 0x24;
		ec_datagram_write(&addrt, &val, 1);
		addrt.addr.position.off = 0x0806 + 16;
		val = 0x01;
		ec_datagram_write(&addrt, &val, 1);

	}
	{
		/* sm3 rx pdo (slave->master) @0x1a00 */
		struct ec_dgram_addr addrt;
		uint16_t val;
		
		addrt.type = EC_DGRAM_ADDR_AUTO_INC;
		addrt.addr.position.pos = 0x0000;

		addrt.addr.position.off = 0x0806 + 24;
		val = 0x00;
		ec_datagram_write(&addrt, &val, 1);

		addrt.addr.position.off = 0x0800 + 24;
		val = 0x1a00;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0802 + 24;
		val = 0x0005;
		ec_datagram_write(&addrt, &val, 2);
		addrt.addr.position.off = 0x0804 + 24;
		val = 0x20;
		ec_datagram_write(&addrt, &val, 1);
		addrt.addr.position.off = 0x0806 + 24;
		val = 0x01;
		ec_datagram_write(&addrt, &val, 1);

	}


	/*****************************************************************
		state test
	 *****************************************************************/

	printf("\n----------\nstate test\n----------\n");

	addr.type = EC_DGRAM_ADDR_AUTO_INC;
	addr.addr.position.pos = htole16(0x0000);
	addr.addr.position.off = htole16(0x0000);

	printf("esc_al_state_get(): %d\n", esc_al_state_get(&addr));

	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("esc_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}
	
	printf("esc_al_state_set(2==ESC_AL_STATE_PRE_OPERATIONAL): %d\n",
	    esc_al_state_set(&addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout));

	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("esc_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}

	printf("ec_al_state_set(4==ESC_AL_STATE_SAFE_OPERATIONAL): %d\n",
	    esc_al_state_set(&addr, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout));

	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("ec_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}

	printf("esc_al_state_set(8==ESC_AL_STATE_OPERATIONAL): %d\n",
	    esc_al_state_set(&addr, ESC_AL_STATE_OPERATIONAL, &timeout));

	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("esc_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}
	/*****************************************************************
		mailbox test
	 *****************************************************************/
	printf("\n----------\nmailbox test\n----------\n");

	addr.type = EC_DGRAM_ADDR_AUTO_INC;
	addr.addr.position.pos = htole16(0x0000);
	addr.addr.position.off = htole16(0x0000);

	mb = esc_mailbox_create(&addr, NULL, NULL);
	printf("esc_mailbox_create: %p\n", mb);

	canopen_dev = esc_coe_create(mb);
	printf("esc_coe_create: %p\n", canopen_dev);

	ret = esc_al_status_code_get(&addr);
	printf("esc_al_status_code_get(): 0x%x\n", ret);
	if (ret != 0) {
		printf("esc_al_error_ack(): %d\n", esc_al_error_ack(&addr));
		ret = esc_al_status_code_get(&addr);
		printf("esc_al_status_code_get(): 0x%x\n", ret);
	}

	/*****************************************************************
		CoE test
	 *****************************************************************/
	printf("\n----------\nCoE test\n----------\n");

	ret = canopen_read_param(canopen_dev, 0x2899, 0x0, buffer, 4);
	printf("canopen_read_param(,0x2899, 0x0) E153:%d\n", ret);
	if (ret == 4)
		printf("param data: 0x%02x 0x%02x 0x%02x 0x%02x\n",
		    buffer[0], buffer[1], buffer[2], buffer[3]);

	ret = canopen_read_param(canopen_dev, 0x289a, 0x0, buffer, 4);
	printf("canopen_read_param(,0x289a, 0x0) E154:%d\n", ret);
	if (ret == 4)
		printf("param data: 0x%02x 0x%02x 0x%02x 0x%02x\n",
		    buffer[0], buffer[1], buffer[2], buffer[3]);

	ret = canopen_read_param(canopen_dev, 0x289b, 0x0, buffer, 4);
	printf("canopen_read_param(,0x289b, 0x0) E155:%d\n", ret);
	if (ret == 4)
		printf("param data: 0x%02x 0x%02x 0x%02x 0x%02x\n",
		    buffer[0], buffer[1], buffer[2], buffer[3]);

	ret = canopen_read_param(canopen_dev, 0x289c, 0x0, buffer, 4);
	printf("canopen_read_param(,0x289c, 0x0) E156:%d\n", ret);
	if (ret == 4)
		printf("param data: 0x%02x 0x%02x 0x%02x 0x%02x\n",
		    buffer[0], buffer[1], buffer[2], buffer[3]);

	ret = canopen_read_param(canopen_dev, 0x2000, 0x0, buffer, 4);
	printf("canopen_read_param(,0x2000, 0x0) A000:%d\n", ret);
	if (ret == 4)
		printf("param data: 0x%02x 0x%02x 0x%02x 0x%02x\n",
		    buffer[0], buffer[1], buffer[2], buffer[3]);

	ret = canopen_read_param(canopen_dev, 0x30de, 0x0, buffer, 4);
	printf("canopen_read_param(,0x30de, 0x0) I222:%d\n", ret);
	if (ret == 4)
		printf("param data: 0x%02x 0x%02x 0x%02x 0x%02x\n",
		    buffer[0], buffer[1], buffer[2], buffer[3]);
t = time(NULL);
i = 0;
while(1) {

	if (time(NULL) != t) {
		printf("\t\t%d / sec\n", i);
		i = 0;
		t = time(NULL);
	}
//	i++;
	
//	ret = canopen_read_param(canopen_dev, 0x3050, 0x0, buffer, 4);
//	printf("canopen_read_param(,0x3050, 0x0) I80:%d\n", ret);
//	if (ret == 4)
//		printf("param data: 0x%02x 0x%02x 0x%02x 0x%02x\n",
//		    buffer[0], buffer[1], buffer[2], buffer[3]);
//		printf("%d\n", le32toh(*(uint32_t*)buffer));


	
	addr.type = EC_DGRAM_ADDR_AUTO_INC;
	addr.addr.position.pos = 0x0000;
	
	addr.addr.position.off = 0x1600;

	buffer[0] = 1;
	buffer[1] = 4;
	*((uint16_t*)(buffer+2)) = htole16(0x0000);
	*((uint16_t*)(buffer+8)) = htole16(speed);
	ret = ec_datagram_write(&addr, buffer, 10);
	if (ret == 10)
		i++;
//	printf("%d %d\n", ret, speed);

	buffer[0] = 1;
	buffer[1] = 4;
	*((uint16_t*)(buffer+2)) = htole16(0x0001);
	*((uint16_t*)(buffer+8)) = htole16(speed);
	ret = ec_datagram_write(&addr, buffer, 10);

	speed+=1;
	if (speed > 16384)
		speed = 8000;

	if (ret == 10)
		i++;
//	printf("%d\n", ret);

	usleep (4000);

	addr.type = EC_DGRAM_ADDR_AUTO_INC;
	addr.addr.position.pos = 0x0000;
	
	addr.addr.position.off = 0x1a00;
	ret = ec_datagram_read(&addr, buffer, 5);

//	for (i = 0; i < 5; i++) {
//		printf("%02x", buffer[i]);
//	}
//	if (le32toh(*(uint32_t*)&buffer[1]) != prev) {
//		prev = le32toh(*(uint32_t*)&buffer[1]);
//		fprintf(stderr, "%d %d\n", speed, le32toh(*(uint32_t*)&buffer[1]));
//		i++;
//	}
//	printf(" %d\n", ret);
}
	return 0;
}
