#include "eth.h"
#include "ec_int.h"
#include "log.h"

#include <net/if.h>
#include <arpa/inet.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <string.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <errno.h>

static struct sockaddr_ll sockaddr_ll;

int eth_timeout_set(int ec_sock, struct timeval *tx, struct timeval *rx)
{
	setsockopt(ec_sock, SOL_SOCKET, SO_SNDTIMEO, tx, sizeof(*tx));
	setsockopt(ec_sock, SOL_SOCKET, SO_RCVTIMEO, rx, sizeof(*rx));

	return 0;
}


int eth_open(char *ifname)
{
	int ec_sock;
	struct ifreq ifreq;
	int ret;
	struct timeval timeout;
	
	ec_sock = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_ETHERCAT));
	if (ec_sock < 0)
		return ec_sock;
	
	strncpy(ifreq.ifr_name, ifname, IFNAMSIZ);
	ret = ioctl(ec_sock, SIOCGIFINDEX, &ifreq);
	if (ret < 0) {
		log_send(LOG_T_ERROR, "ioctl SIOCGIFINDEX: %s", strerror(errno));
		goto err_socket;
	}
	log_send(LOG_T_DEBUG, "Interface %s index nr found: %d", 
	    ifname, ifreq.ifr_ifindex);

	sockaddr_ll.sll_family = AF_PACKET;
	sockaddr_ll.sll_protocol = htons(ETH_P_ETHERCAT);
	sockaddr_ll.sll_ifindex = ifreq.ifr_ifindex;

	ret=bind(ec_sock, (struct sockaddr*)&sockaddr_ll, sizeof(sockaddr_ll));
	if (ret < 0) {
		log_send(LOG_T_ERROR, "bind(): %s", strerror(errno));
		goto err_socket;
	}
	log_send(LOG_T_DEBUG, "Bind to interface and protocol 0x%04x done",
	    ETH_P_ETHERCAT);


	sockaddr_ll.sll_addr[0] = 0xff;
	sockaddr_ll.sll_addr[1] = 0xff;
	sockaddr_ll.sll_addr[2] = 0xff;
	sockaddr_ll.sll_addr[3] = 0xff;
	sockaddr_ll.sll_addr[4] = 0xff;
	sockaddr_ll.sll_addr[5] = 0xff;
	sockaddr_ll.sll_halen = ETH_ALEN;

	/* Using select: set to non-blocking (1)
	   Using SO_SNDTIMEO: set to blocking (0) */
	ioctl(ec_sock, FIONBIO, &(int){0});

	timeout.tv_sec = 0;
	timeout.tv_usec = EC_DGRAM_TIMEOUT;
	setsockopt(ec_sock, SOL_SOCKET, SO_SNDTIMEO, &timeout, sizeof(timeout));
	timeout.tv_sec = 0;
	timeout.tv_usec = EC_DGRAM_TIMEOUT;
	setsockopt(ec_sock, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));

	return ec_sock;

err_socket:
	close(ec_sock);
	return -1;
}

int eth_close(int ec_sock)
{
	return close(ec_sock);
}

int eth_recv(int ec_sock, void *frame, int max_len)
{
	return recv(ec_sock, frame, max_len, 0);
}

int eth_send(int ec_sock, void *frame, int len)
{
	return sendto(ec_sock, frame, len, 
	    MSG_DONTWAIT, 
	    (struct sockaddr *)&sockaddr_ll, sizeof(sockaddr_ll));
}

